from typing import Optional

from dataclasses import dataclass
from casanova import TabularRecord


@dataclass
class RedditPost(TabularRecord):
    title: str
    url: str
    author: str
    author_text: Optional[str]
    points: str
    raw_number_comments: str
    number_comments: int
    published_date: str
    edited_date: str
    external_link: Optional[str]
    error: str


@dataclass
class RedditComment(TabularRecord):
    comment_url: str
    author: str
    id: str
    parent: Optional[str]
    points: Optional[str]
    published_date: Optional[str]
    edited_date: Optional[str]
    comment: str
    error: Optional[str]


@dataclass
class RedditUserPost(TabularRecord):
    title: str
    url: str
    author_text: str
    points: str
    raw_number_comments: str
    number_comments: int
    published_date: str
    edited_date: str
    external_link: str
    subreddit: str
    error: str


@dataclass
class RedditUserComment(TabularRecord):
    post_title: str
    post_url: str
    post_author: str
    post_subreddit: str
    points: str
    published_date: Optional[str]
    edited_date: Optional[str]
    text: str
    link: str
    comment_url: str
    error: Optional[str]



from minet.twitter.api_client import TwitterAPIClient
from minet.twitter.api_scraper import TwitterAPIScraper, TwitterGuestAPIScraper

__all__ = ["TwitterAPIClient", "TwitterAPIScraper", "TwitterGuestAPIScraper"]



from typing import Optional, List, Tuple

from ural import URLFormatter, urlpathsplit

from minet.bluesky.constants import (
    BLUESKY_HTTP_API_BASE_URL,
    BLUESKY_HTTP_API_ALTERNATE_URL,
)


def parse_post_url(url: str) -> Tuple[str, str]:
    path = urlpathsplit(url)

    did = path[1]
    rkey = path[3]

    return did, rkey


def format_post_at_uri(did: str, rkey: str) -> str:
    return "at://{}/app.bsky.feed.post/{}".format(did, rkey)


class BlueskyHTTPAPIUrlFormatter(URLFormatter):
    BASE_URL = BLUESKY_HTTP_API_BASE_URL

    def create_session(self) -> str:
        return self.format(
            path="com.atproto.server.createSession",
        )

    def refresh_session(self) -> str:
        return self.format(path="com.atproto.server.refreshSession")

    def resolve_handle(self, handle: str) -> str:
        # Handles resolving of special handles based on a different DNS do not work on the regular API, we need to use the alternate endpoint from the public facing API
        # cf https://github.com/bluesky-social/indigo/issues/833
        return self.format(
            base_url=BLUESKY_HTTP_API_ALTERNATE_URL,
            path="com.atproto.identity.resolveHandle",
            args={"handle": handle},
        )

    def get_posts(self, uris: List[str]) -> str:
        return self.format(
            path="app.bsky.feed.getPosts", args=[("uris", uri) for uri in uris]
        )

    def get_profiles(self, identifiers: List[str]) -> str:
        return self.format(
            path="app.bsky.actor.getProfiles",
            args=[("actors", identifier) for identifier in identifiers],
        )

    def search_posts(
        self, q: str, cursor: Optional[str] = None, limit: int = 100
    ) -> str:
        return self.format(
            path="app.bsky.feed.searchPosts",
            args={"q": q, "cursor": cursor, "limit": limit},
        )



from minet.scrape.classes.base import ScraperBase
from minet.scrape.classes.definition import DefinitionScraper, validate, scrape
from minet.scrape.classes.function import FunctionScraper
from minet.scrape.classes.named import NamedScraper, NAMED_SCRAPERS

__all__ = [
    "ScraperBase",
    "DefinitionScraper",
    "validate",
    "scrape",
    "FunctionScraper",
    "NamedScraper",
    "NAMED_SCRAPERS",
]



from typing import Iterator, Optional, List, Dict
from time import time

from minet.web import (
    request,
    create_request_retryer,
    create_pool_manager,
    retrying_method,
)

from minet.tiktok.urls import (
    TiktokHTTPAPIUrlFormatter,
)
from minet.tiktok.types import TiktokCommercialContent
from minet.tiktok.constants import TIKTOK_COMMERCIAL_CONTENTS_MAX_COUNT
from minet.tiktok.exceptions import TiktokAuthenticationError, TiktokHTTPAPIError


class TiktokHTTPClient:
    def __init__(self, identifier: str, password: str):
        self.urls = TiktokHTTPAPIUrlFormatter()
        self.pool_manager = create_pool_manager()
        self.retryer = create_request_retryer()

        # First auth
        self.create_session(identifier, password)

    @retrying_method()
    def create_session(self, key: str, secret: str):
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = {
            "client_key": key,
            "client_secret": secret,
            "grant_type": "client_credentials",
        }

        response = request(
            self.urls.create_session(),
            pool_manager=self.pool_manager,
            method="POST",
            urlencoded_body=data,
            headers=headers,
        )

        if response.status != 200:
            raise TiktokAuthenticationError

        data = response.json()

        self.expired_at = time() + data["expires_in"]
        self.access_token = data["access_token"]

    def is_token_expired(self):
        return self.expired_at - time() < 10

    @retrying_method()
    def request(self, url: str, params: Dict, headers: Dict):
        if self.is_token_expired():
            self.create_session()

        response = request(
            url,
            pool_manager=self.pool_manager,
            method="POST",
            json_body=params,
            known_encoding="utf-8",
            headers=headers,
        )

        return response

    def search_commercial_contents(
        self,
        country: str,
        min_date: str,
        max_date: str,
        usernames: Optional[List[str]] = [],
    ) -> Iterator[Dict]:
        headers = {
            "authorization": "Bearer {}".format(self.access_token),
            "Content-Type": "application/json",
        }

        filters = {
            "content_published_date_range": {"min": min_date, "max": max_date},
            "creator_country_code": country,
        }

        params = {
            "filters": filters,
            "max_count": TIKTOK_COMMERCIAL_CONTENTS_MAX_COUNT,
        }

        if usernames:
            filters["creator_usernames"] = usernames

        url = self.urls.search_commercial_contents()

        while True:
            response = self.request(url, params, headers)

            data = response.json()

            if data["error"]["code"] != "ok":
                raise TiktokHTTPAPIError(data["error"])

            for content in data["data"]["commercial_contents"]:
                yield TiktokCommercialContent.from_payload(content, collected_via="api")

            has_more = data["data"]["has_more"]

            if has_more:
                params["search_id"] = data["data"]["search_id"]
            else:
                break



from typing import Iterator, Iterable, Optional, Any, List, Dict

from time import time, sleep
from ebbe import as_reconciled_chunks

from minet.web import (
    create_request_retryer,
    create_pool_manager,
    retrying_method,
    request,
    Response,
)

from minet.bluesky.urls import (
    BlueskyHTTPAPIUrlFormatter,
    parse_post_url,
    format_post_at_uri,
)
from minet.bluesky.jwt import parse_jwt_for_expiration
from minet.bluesky.types import BlueskyPost
from minet.bluesky.exceptions import (
    BlueskyAuthenticationError,
    BlueskySessionRefreshError,
)


class BlueskyHTTPClient:
    def __init__(self, identifier: str, password: str):
        self.urls = BlueskyHTTPAPIUrlFormatter()
        self.pool_manager = create_pool_manager()
        self.retryer = create_request_retryer()
        self.rate_limit_reset: Optional[int] = None

        # First auth
        self.create_session(identifier, password)

    @retrying_method()
    def create_session(self, identifier: str, password: str):
        response = request(
            self.urls.create_session(),
            pool_manager=self.pool_manager,
            method="POST",
            json_body={"identifier": identifier, "password": password},
        )

        if response.status != 200:
            raise BlueskyAuthenticationError

        data = response.json()

        self.access_jwt = data["accessJwt"]
        self.refresh_jwt = data["refreshJwt"]
        self.access_jwt_expiration = parse_jwt_for_expiration(self.access_jwt)

    def is_access_jwt_expired(self) -> bool:
        # If the token has 10 seconds left, we consider it expired to avoid network-related issues
        return self.access_jwt_expiration - time() < 10

    @retrying_method()
    def refresh_session(self):
        response = request(
            self.urls.refresh_session(),
            pool_manager=self.pool_manager,
            method="POST",
            headers={"Authorization": "Bearer {}".format(self.refresh_jwt)},
        )

        if response.status != 200:
            raise BlueskySessionRefreshError

        data = response.json()

        self.access_jwt = data["accessJwt"]
        self.refresh_jwt = data["refreshJwt"]
        self.access_jwt_expiration = parse_jwt_for_expiration(self.access_jwt)

    @retrying_method()
    def request(
        self,
        url: str,
        method: str = "GET",
        json_body=None,
    ) -> Response:
        if self.rate_limit_reset is not None:
            sleep(self.rate_limit_reset - time() + 0.1)
            self.rate_limit_reset = None

        if self.is_access_jwt_expired():
            self.refresh_session()

        headers = {"Authorization": "Bearer {}".format(self.access_jwt)}

        response = request(
            url,
            pool_manager=self.pool_manager,
            method=method,
            json_body=json_body,
            known_encoding="utf-8",
            headers=headers,
        )

        remaining = int(response.headers["RateLimit-Remaining"])

        if remaining <= 0:
            self.rate_limit_reset = int(response.headers["RateLimit-Reset"])

        return response

    def search_posts(self, query: str) -> Iterator[BlueskyPost]:
        cursor = None

        while True:
            url = self.urls.search_posts(query, cursor=cursor)

            response = self.request(url)
            data = response.json()

            for post in data["posts"]:
                yield BlueskyPost.from_payload(post)

            cursor = data.get("cursor")

            if cursor is None:
                break

    def resolve_handle(self, identifier: str) -> str:
        url = self.urls.resolve_handle(identifier)

        response = self.request(url)
        data = response.json()

        return data["did"]

    def post_url_to_did_at_uri(self, url: str) -> str:
        handle, rkey = parse_post_url(url)
        did = self.resolve_handle(handle)

        return format_post_at_uri(did, rkey)

    # NOTE: this API route that not return any results for at-uris containing handles!
    def get_posts(self, did_at_uris: Iterable[str]) -> Iterator[Any]:
        def work(chunk: List[str]) -> Dict[str, Any]:
            url = self.urls.get_posts(chunk)
            response = self.request(url)
            data = response.json()

            return {post["uri"]: post for post in data["posts"]}

        def reconcile(data: Dict[str, Any], uri: str) -> Any:
            return data.get(uri)

        for _, post_data in as_reconciled_chunks(25, did_at_uris, work, reconcile):
            yield post_data

    def get_profiles(self, identifiers: Iterable[str]) -> Iterator[Any]:
        def work(chunk: List[str]) -> Dict[str, Any]:
            url = self.urls.get_profiles(chunk)
            response = self.request(url)
            data = response.json()

            index = {}

            for profile in data["profiles"]:
                index[profile["did"]] = profile
                index[profile["handle"]] = profile

            return index

        def reconcile(data: Dict[str, Any], identifier: str) -> Any:
            return data.get(identifier)

        for _, profile_data in as_reconciled_chunks(25, identifiers, work, reconcile):
            yield profile_data



from minet.bluesky.exceptions import (
    BlueskyAuthenticationError,
    BlueskySessionRefreshError,
)

from minet.cli.utils import with_fatal_errors


def fatal_errors_hook(_, e):
    if isinstance(e, BlueskyAuthenticationError):
        return [
            "Invalid Bluesky authentication!",
            "Did you give correct information to --identifier (did you forget `.bsky.social` at the end?) and --password flags?",
            "If you don't have a password, you can create one here: [cyan]https://bsky.app/settings/app-passwords",
        ]

    if isinstance(e, BlueskySessionRefreshError):
        return ["Something went wrong when refreshing the auth session.", "Sorry :("]


def with_bluesky_fatal_errors(fn):
    return with_fatal_errors(fatal_errors_hook)(fn)



# =============================================================================
# Minet Facebook CLI Action
# =============================================================================
#
# Logic of the `fb` action.
#
from minet.cli.argparse import command

FACEBOOK_EXPERIMENTAL_COMMENTS_SUBCOMMAND = command(
    "experimental-comments",
    "minet.cli.facebook.experimental_comments",
    title="Minet Experimental Facebook Comments Command",
    description="""
        Scrape a Facebook post's comments using an experimental scraper
        based on browser emulation.
    """,
    epilog="""
        Examples:

        . Scraping a post's comments:
            $ minet fb experimental-comments https://www.facebook.com/groups/186982532676569/permalink/4096995827030341/ > comments.csv

        . Scraping comments from multiple posts listed in a CSV file:s
            $ minet fb experimental-comments post_url -i posts.csv > comments.csv
    """,
    variadic_input={"dummy_column": "post_url", "item_label": "post url"},
    arguments=[
        {
            "flag": "--headful",
            "help": "Whether to disable the browser's headless mode for debugging purposes.",
            "action": "store_true",
        }
    ],
)

FACEBOOK_URL_LIKES_SUBCOMMAND = command(
    "url-likes",
    "minet.cli.facebook.url_likes",
    title="Minet Facebook Url Likes Command",
    description="""
        Retrieve the approximate number of "likes" (actually an aggregated engagement metric)
        that a url got on Facebook. The command can also be used with a list of urls stored in a CSV file.
        This number is found by scraping Facebook's share button, which only gives a
        rough estimation of the real engagement metric: "Share 45K" for example.

        Note that this number does not actually only correspond to the number of
        likes or shares, but it is rather the sum of like, love, ahah, angry, etc.
        reactions plus the number of comments and shares that the URL got on Facebook
        (here is the official documentation: https://developers.facebook.com/docs/plugins/faqs
        explaining "What makes up the number shown next to my Share button?").
    """,
    epilog="""
        example:
        . Retrieving the "like" number for one url:
            $ minet fb url-likes "www.example-url.com" > url_like.csv

        . Retrieving the "like" number for the urls listed in a CSV file:
            $ minet fb url-likes url -i url.csv > url_likes.csv
    """,
    variadic_input={"dummy_column": "url", "item_label": "url"},
)

FACEBOOK_COMMAND = command(
    "facebook",
    "minet.cli.facebook",
    "Minet Facebook Command",
    aliases=["fb"],
    description="""
        Collect data from Facebook.
    """,
    subcommands=[
        FACEBOOK_EXPERIMENTAL_COMMENTS_SUBCOMMAND,
        FACEBOOK_URL_LIKES_SUBCOMMAND,
    ],
)



# =============================================================================
# Minet Reddit CLI Action
# =============================================================================
#
# Logic of the `rd` action.
#

from minet.cli.argparse import command

REDDIT_POSTS_SUBCOMMAND = command(
    "posts",
    "minet.cli.reddit.posts",
    title="Minet Reddit Posts Command",
    description="""
        Retrieve reddit posts from a subreddit link or name.
    """,
    epilog="""
        Example:

        . Searching posts from the subreddit r/france:
            $ minet reddit posts https://www.reddit.com/r/france > r_france_posts.csv
            $ minet reddit posts france > r_france_posts.csv
            $ minet reddit posts r/france > r_france_posts.csv
    """,
    variadic_input={
        "dummy_column": "subreddit",
        "item_label": "subreddit url, shortcode or id",
        "item_label_plural": "subreddit urls, shortcodes or ids",
    },
    arguments=[
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of posts to retrieve.",
            "type": int,
        },
        {
            "flags": ["-t", "--text"],
            "help": "Retrieve the text of the post. Note that it will require one request per post.",
            "action": "store_true",
        },
    ],
)

REDDIT_COMMENTS_SUBCOMMAND = command(
    "comments",
    "minet.cli.reddit.comments",
    title="Minet Reddit Comments Command",
    description="""
        Retrieve comments from a reddit post link.
        Note that it will only retrieve the comments displayed on the page. If you want all the comments you need to use -A, --all but it will require a request per comment, and you can only make 100 requests per 10 minutes.
    """,
    epilog="""
        Example:

        . Searching comments from a reddit post:
            $ minet reddit comments https://www.reddit.com/r/france/comments/... > r_france_comments.csv
    """,
    variadic_input={
        "dummy_column": "post",
        "item_label": "post url, shortcode or id",
        "item_label_plural": "posts urls, shortcodes or ids",
    },
    arguments=[
        {
            "flags": ["-A", "--all"],
            "help": "Retrieve all comments.",
            "action": "store_true",
        },
    ],
)

REDDIT_USER_POSTS_SUBCOMMAND = command(
    "user-posts",
    "minet.cli.reddit.user_posts",
    title="Minet Reddit User Posts Command",
    description="""
        Retrieve reddit posts from a user link.
    """,
    epilog="""
        Example:

        . Searching posts from the user page of u/random_user:
            $ minet reddit user-posts https://www.reddit.com/user/random_user/submitted/ > random_user_posts.csv
    """,
    variadic_input={
        "dummy_column": "user",
        "item_label": "user url, shortcode or id",
        "item_label_plural": "user urls, shortcodes or ids",
    },
    arguments=[
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of posts to retrieve.",
            "type": int,
        },
        {
            "flags": ["-t", "--text"],
            "help": "Retrieve the text of the post. Note that it will require one request per post.",
            "action": "store_true",
        },
    ],
)

REDDIT_USER_COMMENTS_SUBCOMMAND = command(
    "user-comments",
    "minet.cli.reddit.user_comments",
    title="Minet Reddit User Comments Command",
    description="""
        Retrieve reddit comments from a user link.
    """,
    epilog="""
        Example:

        . Searching comments from the user page of u/random_user:
            $ minet reddit user-comments https://www.reddit.com/user/random_user/comments/ > random_user_comments.csv
    """,
    variadic_input={
        "dummy_column": "user",
        "item_label": "user url, shortcode or id",
        "item_label_plural": "user urls, shortcodes or ids",
    },
    arguments=[
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of comments to retrieve.",
            "type": int,
        },
    ],
)

REDDIT_COMMAND = command(
    "reddit",
    "minet.cli.reddit",
    "Minet Reddit Command",
    aliases=["rd"],
    description="""
        Collect data from Reddit.
    """,
    subcommands=[
        REDDIT_POSTS_SUBCOMMAND,
        REDDIT_COMMENTS_SUBCOMMAND,
        REDDIT_USER_POSTS_SUBCOMMAND,
        REDDIT_USER_COMMENTS_SUBCOMMAND,
    ],
)



# =============================================================================
# Minet Tiktok Constants
# =============================================================================
#
# General constants used throughout the Tiktok functions.
#
from urllib3 import Timeout

TIKTOK_URL = "https://www.tiktok.com"
TIKTOK_PUBLIC_API_DEFAULT_TIMEOUT = Timeout(connect=10, read=60 * 2)
TIKTOK_DEFAULT_THROTTLE = 3.0
TIKTOK_MIN_TIME_RETRYER = 5.0
TIKTOK_MAX_RANDOM_ADDENDUM = 1.0

TIKTOK_HTTP_API_BASE_URL = "https://open.tiktokapis.com/v2"
TIKTOK_COMMERCIAL_CONTENTS_FIELDS = (
    "id,create_timestamp,create_date,label,brand_names,creator,videos"
)
TIKTOK_COMMERCIAL_CONTENTS_MAX_COUNT = 50

TIKTOK_COMMERCIAL_CONTENTS_LABELS = {
    "1": "Paid Partnership",
    "2": "Promotional Content",
}



import libipld

from minet.cli.console import console

from minet.bluesky.websocket_client import BlueskyWebSocketClient


def action(_):
    client = BlueskyWebSocketClient()

    with client.subscribe_repos() as socket:
        while True:
            payload = socket.recv()
            header, _ = libipld.decode_dag_cbor_multi(payload)  # type: ignore
            console.print(header, highlight=True)



from ural import URLFormatter

from minet.tiktok.constants import (
    TIKTOK_HTTP_API_BASE_URL,
    TIKTOK_COMMERCIAL_CONTENTS_FIELDS,
)


class TiktokHTTPAPIUrlFormatter(URLFormatter):
    BASE_URL = TIKTOK_HTTP_API_BASE_URL

    def create_session(self) -> str:
        return self.format(
            path="oauth/token/",
        )

    def search_commercial_contents(self) -> str:
        return self.format(
            path="research/adlib/commercial_content/query/",
            args={"fields": TIKTOK_COMMERCIAL_CONTENTS_FIELDS},
        )



import re
from typing import List, Optional, Union, TypeVar, cast, overload

import warnings
from contextlib import contextmanager
from bs4 import Tag, BeautifulSoup, SoupStrainer

from minet.scrape.std import get_display_text

try:
    from bs4 import XMLParsedAsHTMLWarning
except ImportError:
    XMLParsedAsHTMLWarning = None

WHITESPACE_RE = re.compile(r"\s+")

T = TypeVar("T")


class SelectionError(Exception):
    pass


class ExtractionError(Exception):
    pass


def normalize_css(css: str) -> str:
    return css.replace(":contains(", ":-soup-contains(")


def extract(elem: "MinetTag", target: Optional[str]) -> Optional[str]:
    if target is None or target == "text":
        return elem.get_text()

    if target == "html" or target == "inner_html":
        return elem.get_html()

    if target == "outer_html":
        return elem.get_outer_html()

    if target == "display_text":
        return elem.get_display_text()

    return cast(Optional[str], elem.get(target))


class MinetTag(Tag):
    def force_select_one(self, css: str, *args, **kwargs) -> "MinetTag":
        css = normalize_css(css)

        elem = super().select_one(css, *args, **kwargs)

        if elem is None:
            raise SelectionError(css)

        return cast(MinetTag, elem)

    def select_one(self, css: str, *args, **kwargs) -> Optional["MinetTag"]:
        css = normalize_css(css)
        return cast(Optional["MinetTag"], super().select_one(css, *args, **kwargs))

    def select(self, css: str, *args, **kwargs) -> List["MinetTag"]:
        css = normalize_css(css)
        return cast(List[MinetTag], super().select(css, *args, **kwargs))

    def find(
        self,
        name: Optional[str] = None,
        attrs={},
        recursive: bool = True,
        string=None,
        **kwargs,
    ) -> Optional["MinetTag"]:
        return cast(
            Optional[MinetTag],
            super().find(name, attrs, recursive, string, **kwargs),
        )

    def force_find(
        self,
        name: Optional[str] = None,
        attrs={},
        recursive: bool = True,
        string=None,
        **kwargs,
    ) -> "MinetTag":
        elem = super().find(name, attrs, recursive, string, **kwargs)

        if elem is None:
            raise SelectionError

        return cast(MinetTag, elem)

    def find_all(
        self,
        name: Optional[str] = None,
        attrs={},
        recursive: bool = True,
        string=None,
        limit: Optional[int] = None,
        **kwargs,
    ) -> List["MinetTag"]:
        return cast(
            List["MinetTag"],
            super().find_all(name, attrs, recursive, string, limit, **kwargs),
        )

    def force_scrape_one(self, css: str, target: Optional[str] = None) -> str:
        elem = self.select_one(css)

        if elem is None:
            raise SelectionError(css)

        value = extract(elem, target)

        if value is None:
            raise ExtractionError(target)

        return value

    def scrape_one(self, css: str, target: Optional[str] = None) -> Optional[str]:
        elem = self.select_one(css)

        if elem is None:
            return None

        return extract(elem, target)

    def scrape(self, css: str, target: Optional[str] = None) -> List[str]:
        output = []

        for elem in self.select(css):
            value = extract(elem, target)

            if value is not None:
                output.append(value)

        return output

    def get_text(self) -> str:
        return super().get_text().strip()

    def get_display_text(self) -> str:
        return get_display_text(self)

    def get_html(self) -> str:
        return self.decode_contents().strip()

    def get_inner_html(self) -> str:
        return self.get_html()

    def get_outer_html(self) -> str:
        return str(self).strip()

    def __getitem__(self, name: str) -> str:
        return cast(str, super().__getitem__(name))

    @overload
    def get(self, name: str, default: None = ...) -> Optional[str]: ...

    @overload
    def get(self, name: str, default: T = ...) -> Union[T, str]: ...

    def get(self, name: str, default: Optional[T] = None) -> Optional[Union[T, str]]:
        return cast(Optional[str], super().get(name, default))

    def get_list(self, name: str) -> List[str]:
        value = super().get(name)

        if not value:
            return []

        return WHITESPACE_RE.split(cast(str, value).strip())


WONDERFUL_ELEMENT_CLASSES = {Tag: MinetTag}


class WonderfulSoup(BeautifulSoup, MinetTag):
    """
    A subclass of BeautifulSoup with some handy tricks such as selection
    wrangling, scraping utilities and better typings.
    """

    def __init__(
        self,
        markup: str,
        features: str = "html.parser",
        parse_only: Optional[SoupStrainer] = None,
    ) -> None:
        super().__init__(
            markup,
            features,
            element_classes=WONDERFUL_ELEMENT_CLASSES,  # type: ignore
            parse_only=parse_only,
            multi_valued_attributes=None,
        )


@contextmanager
def suppress_xml_parsed_as_html_warnings(bypass=False):
    try:
        if XMLParsedAsHTMLWarning is None or bypass:
            yield
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=XMLParsedAsHTMLWarning)
                yield
    finally:
        pass



from minet.cli.argparse import command, ConfigAction

BLUESKY_HTTP_API_COMMON_ARGUMENTS = [
    {
        "flag": "--identifier",
        "help": "Bluesky personal identifier (don't forget the `.bsky.social` at the end).",
        "rc_key": ["bluesky", "identifier"],
        "action": ConfigAction,
    },
    {
        "flag": "--password",
        "help": "Bluesky app password (not your personal password, must be created here: https://bsky.app/settings/app-passwords).",
        "rc_key": ["bluesky", "password"],
        "action": ConfigAction,
    },
]

BLUESKY_FIREHOSE_COMMAND = command(
    "firehose",
    "minet.cli.bluesky.firehose",
    title="Minet Bluesky Firehose Command",
    description="""
        Plug into the Bluesky Firehose.
    """,
)

BLUESKY_SEARCH_POSTS_COMMAND = command(
    "search-posts",
    "minet.cli.bluesky.search_posts",
    title="Minet Bluesky Search Post Command",
    description="""
        Search for Bluesky posts using their HTTP API.
    """,
    variadic_input={"dummy_column": "query"},
    arguments=[*BLUESKY_HTTP_API_COMMON_ARGUMENTS],
)

BLUESKY_COMMAND = command(
    "bluesky",
    "minet.cli.bluesky",
    aliases=["bsky"],
    title="Minet Bluesky Command",
    description="""
        Collect data from Bluesky.
    """,
    subcommands=[BLUESKY_FIREHOSE_COMMAND, BLUESKY_SEARCH_POSTS_COMMAND],
)



from minet.crawl.types import (
    CrawlResult,
    AnyCrawlResult,
    SuccessfulCrawlResult,
    ErroredCrawlResult,
    CrawlJob,
    CrawlTarget,
    UrlOrCrawlTarget,
)
from minet.crawl.state import CrawlerState
from minet.crawl.spiders import (
    Spider,
    SpiderResult,
    SpiderNextTargets,
    FunctionSpider,
    BasicSpider,
)
from minet.crawl.crawler import Crawler, AnySpider, SpiderDeclaration

__all__ = [
    "CrawlResult",
    "AnyCrawlResult",
    "SuccessfulCrawlResult",
    "ErroredCrawlResult",
    "CrawlJob",
    "CrawlTarget",
    "UrlOrCrawlTarget",
    "CrawlerState",
    "Spider",
    "SpiderResult",
    "SpiderNextTargets",
    "FunctionSpider",
    "BasicSpider",
    "Crawler",
    "AnySpider",
    "SpiderDeclaration",
]



# =============================================================================
# Minet Tiktok Search Commercial Contents CLI Action
# =============================================================================
#
# Logic of the `tiktok search-commercials` action.
#
from itertools import islice
import casanova

from minet.cli.utils import with_loading_bar
from minet.tiktok import TiktokHTTPClient
from minet.tiktok.types import TiktokCommercialContent


@with_loading_bar(
    title="Searching commercials",
    unit="commercials",
    nested=False,
)
def action(cli_args, loading_bar):
    client = TiktokHTTPClient(identifier=cli_args.key, password=cli_args.secret)

    generator = client.search_commercial_contents(
        country=cli_args.country, min_date=cli_args.min_date, max_date=cli_args.max_date
    )

    writer = casanova.writer(cli_args.output, fieldnames=TiktokCommercialContent)

    if cli_args.total:
        generator = islice(generator, cli_args.total)

    for video in generator:
        with loading_bar.step(video):
            writer.writerow(video.as_csv_row())



# Extracted from twitwi
import json
from time import sleep, time
from operator import itemgetter
from twitter import Twitter, OAuth, OAuth2, TwitterHTTPError, Twitter2

from minet.twitter.exceptions import TwitterWrapperMaxAttemptsExceeded
from minet.twitter.constants import APP_ONLY_ROUTES

DEFAULT_MAX_ATTEMPTS = 5

# Established from: https://developer.twitter.com/en/support/twitter-api/error-troubleshooting
NO_RETRY_STATUSES = set([400, 401, 403, 404, 406, 410, 422])


class TwitterWrapper(object):
    def __init__(
        self,
        token,
        token_secret,
        consumer_key,
        consumer_secret,
        listener=None,
        api_version="1.1",
    ):
        api_version = str(api_version)

        if api_version not in ["1.1", "2"]:
            raise TypeError("API version can only be '1.1' or '2'.")

        self.oauth = OAuth(token, token_secret, consumer_key, consumer_secret)

        bearer_token_client = Twitter(
            api_version=None,
            format="",
            secure=True,
            auth=OAuth2(consumer_key, consumer_secret),
        )

        bearer_token = json.loads(
            bearer_token_client.oauth2.token(grant_type="client_credentials")
        )["access_token"]

        self.oauth2 = OAuth2(bearer_token=bearer_token)

        self.auth = {}
        self.waits = {}

        TwitterClass = Twitter

        if api_version == "2":
            TwitterClass = Twitter2

            for route in APP_ONLY_ROUTES:
                self.auth[route] = "app"
                self.waits[route] = {"app": 0}

        self.endpoints = {
            "user": TwitterClass(auth=self.oauth),
            "app": TwitterClass(auth=self.oauth2),
        }

        self.listener = listener

    def call(self, route, max_attempts=DEFAULT_MAX_ATTEMPTS, **kwargs):
        attempts = 0

        if not isinstance(route, list):
            raise TypeError(
                'twitwi.TwitterWrapper.call: expecting route as a list, such as ["friends", "ids"].'
            )

        route = "/".join(route)

        while attempts < max_attempts:
            if route not in self.auth:
                self.auth[route] = "user"

            auth = self.auth[route]

            try:
                return self.endpoints[auth].__getattr__(route)(**kwargs)

            except TwitterHTTPError as e:
                # Rate limited
                if e.e.code == 429:
                    now = int(time())

                    # If there are still API calls available, we obviously
                    # queried Twitter too fast and should just let it breathe
                    remaining = int(e.e.headers["X-Rate-Limit-Remaining"])
                    if remaining > 0:
                        attempts += 1

                        if callable(self.listener):
                            self.listener(
                                "excessive-rate",
                                {
                                    "error": e,
                                    "route": route,
                                    "attempts": attempts,
                                    "auth": auth,
                                },
                            )

                        sleep(1)
                        continue

                    reset = int(e.e.headers["X-Rate-Limit-Reset"])

                    if route not in self.waits:
                        self.waits[route] = {"user": now, "app": now}

                    self.waits[route][auth] = reset

                    if callable(self.listener):
                        self.listener(
                            "rate-limited",
                            {
                                "route": route,
                                "kwargs": kwargs,
                                "reset": reset,
                                "auth": auth,
                            },
                        )

                    min_wait = min(self.waits[route].items(), key=itemgetter(1))

                    if min_wait[1] > now:
                        sleeptime = 5 + max(0, int(min_wait[1] - now))

                        if callable(self.listener):
                            self.listener(
                                "waiting",
                                {
                                    "auth": min_wait[0],
                                    "reset": min_wait[1],
                                    "sleep": sleeptime,
                                },
                            )

                        sleep(sleeptime)

                    self.auth[route] = min_wait[0]

                    continue

                # Errors that should terminate immediately
                elif e.e.code in NO_RETRY_STATUSES:
                    raise e

                # Different error
                else:
                    attempts += 1

                    if callable(self.listener):
                        self.listener("error", {"error": e})

        raise TwitterWrapperMaxAttemptsExceeded
